"""Module for Git Pull Request Threads models."""

from datetime import datetime
from typing import Annotated, Any

from pydantic import BeforeValidator, Field

from ...base.models import BaseCollection, JSONModel
from ...core.models import Properties
from .enums import CommentTypeEnum, ThreadStatusEnum


class PullRequestThreadCommentCreate(JSONModel):
    """Model for creating a new comment on a pull request thread."""

    content: str
    comment_type: Annotated[CommentTypeEnum, BeforeValidator(CommentTypeEnum.validate), Field(alias="commentType")] = (
        CommentTypeEnum.system
    )


class PullRequestThreadComment(JSONModel):
    """Model for a comment on a pull request thread."""

    id: int
    parent_comment_id: Annotated[int, Field(alias="parentCommentId")]
    published_date: Annotated[datetime, Field(alias="publishedDate")]
    last_updated_date: Annotated[datetime, Field(alias="lastUpdatedDate")]
    last_content_updated_date: Annotated[datetime, Field(alias="lastContentUpdatedDate")]
    content: str | None = None
    comment_type: Annotated[CommentTypeEnum, BeforeValidator(CommentTypeEnum.validate), Field(alias="commentType")]


class PullRequestThreadCreate(JSONModel):
    """Model for creating a new pull request thread."""

    comments: list[PullRequestThreadCommentCreate]
    status: Annotated[ThreadStatusEnum | None, BeforeValidator(ThreadStatusEnum.validate)] = None
    properties: dict[str, Any] = {}


class PullRequestThread(JSONModel):
    """Model for a pull request thread."""

    id: int
    published_date: Annotated[datetime, Field(alias="publishedDate")]
    last_updated_date: Annotated[datetime, Field(alias="lastUpdatedDate")]
    deleted: Annotated[bool, Field(alias="isDeleted")]
    comments: list[PullRequestThreadComment]
    properties: dict[str, Properties] | None = None


class PullRequestThreadCollection(BaseCollection[PullRequestThread]):
    """Represent a collection of git pull request threads reference."""



"""Base class for special enums."""

from enum import Enum


class StrIntEnum(int, Enum):
    """Base class for enum that match string to int (vice-versa)."""

    @classmethod
    def validate(cls, value: str | int):
        """Value validator."""
        try:
            if isinstance(value, str):
                return cls[value]
            elif isinstance(value, int):
                return cls(value)
            else:
                raise ValueError(value)
        except (KeyError, ValueError) as e:
            raise ValueError(f"Error: {e}. Valid choices are: {[f'{e.name}={e.value}' for e in cls]}")



"""Test Base models."""

import pytest

from ez_ados.base.models import BaseCollection, JSONModel


# Test JSONModel
def test_jsonmodel_serialization():
    """Test serialization behavior with aliasing."""
    model = JSONModel()
    assert model.model_dump() == {}


# Test BaseCollection
def test_basecollection_count():
    """Test if BaseCollection count is correctly updated."""
    collection = BaseCollection[str]()
    collection.append("test")
    assert collection.count == 1
    collection.append("second")
    assert collection.count == 2


def test_basecollection_iteration():
    """Test if BaseCollection is a list."""
    collection = BaseCollection[str]()
    collection.append("a")
    collection.append("b")
    assert list(collection) == ["a", "b"]


def test_basecollection_getitem():
    """Test fetching an item from a BaseCollection."""
    collection = BaseCollection[str]()
    collection.append("item")
    assert collection[0] == "item"


def test_basecollection_validator():
    """BaseCollection should fail when count doesn't match length."""
    with pytest.raises(ValueError):
        BaseCollection[str](value="test")



"""Models for the Core Azure DevOps API endpoint."""

from typing import Annotated, Any

from pydantic import Field

from ..base.models import JSONModel


class Project(JSONModel):
    """Model for a project."""

    id: str
    name: str


class Properties(JSONModel):
    """Model for properties attached to resources."""

    type: Annotated[str | None, Field(alias="$type", default=None)] = None
    value: Annotated[Any, Field(alias="$value")]



"""Module for base class models."""

from typing import Annotated, Self

from pydantic import BaseModel, ConfigDict, Field, computed_field, model_validator


class JSONModel(BaseModel):
    """Base class for a model serializable to JSON."""

    model_config = ConfigDict(
        serialize_by_alias=True,
        validate_by_name=True,
        validate_by_alias=True,
    )


class BaseCollection[T](JSONModel):
    """Base class for a collection of resources in Azure DevOps."""

    value: Annotated[list[T], Field(default=[])] = []

    @computed_field  # type: ignore[prop-decorator]
    @property
    def count(self) -> int:
        """Return the number of items in this collection."""
        return len(self.value)

    def append(self, element: T):
        """Append an element to the collection."""
        self.value.append(element)

    def __iter__(self):
        """Return an iterator on all values."""
        return self.value.__iter__()

    def __getitem__(self, index: int):
        """Return a value from given index."""
        return self.value[index]



"""Module for Pipeline client API."""

import logging

from pathlib import PurePosixPath, PureWindowsPath

from ..base.clients import Client
from .models import Pipeline, PipelineCollection, PipelineConfiguration, PipelineConfigurationRepository, PipelineCreate

# Get a logger for this module
logger = logging.getLogger(__name__)


class PipelineClient(Client):
    """Represent a client to Pipelines API in Azure DevOps."""

    def get(self, id: int) -> Pipeline:
        """Get a pipeline by ID."""
        return Pipeline.model_validate(self._client.get(str(id)).raise_for_status().json())

    def list(self) -> PipelineCollection:
        """List all pipelines available in a project."""
        return PipelineCollection.model_validate(self._client.get("").raise_for_status().json())

    def create(
        self, name: str, folder: str | PureWindowsPath, yaml_path: str | PurePosixPath, yaml_repository_id: str
    ) -> Pipeline:
        """Create a new pipeline in a project."""
        logger.info("Creating pipeline '%s/%s'...", folder, name)
        if not isinstance(folder, PureWindowsPath):
            _folder = PureWindowsPath(folder)
        else:
            _folder = folder

        if not isinstance(yaml_path, PurePosixPath):
            _yaml_path = PurePosixPath(yaml_path)
        else:
            _yaml_path = yaml_path

        request_body = PipelineCreate(
            name=name,
            folder=_folder,
            configuration=PipelineConfiguration(
                path=_yaml_path,
                repository=PipelineConfigurationRepository(id=yaml_repository_id),
            ),
        )
        logger.debug("Post request body: %s", request_body.model_dump(mode="json"))
        response = self._client.post("", json=request_body.model_dump(mode="json")).raise_for_status()
        return Pipeline.model_validate_json(response.content)



"""Module for Builds client API."""

import logging

from ..base.clients import Client
from ..git.models import GitRepository
from .models import (
    BuildDefinition,
    BuildDefinitionCollection,
    BuildDefinitionCreate,
)

# Get a logger for this module
logger = logging.getLogger(__name__)


class BuildClient(Client):
    """Represent a client to Builds API in Azure DevOps."""

    def get_build_definition(self, id: str) -> BuildDefinition:
        """Get a build definition by ID."""
        logger.debug("Getting build definition id=%d", id)
        response = BuildDefinition.model_validate(self._client.get(f"/definitions/{id}").raise_for_status().json())
        logger.debug("Build definition id=%s specs: %s", id, response)
        return response

    def list_build_definitions(self, repository: GitRepository | None = None) -> BuildDefinitionCollection:
        """List all build definitions available in a project."""
        _query_params: dict[str, str] = {}

        # Build query params
        if isinstance(repository, GitRepository):
            _query_params["repositoryId"] = repository.id
            _query_params["repositoryType"] = "TfsGit"

        return BuildDefinitionCollection.model_validate(
            self._client.get("/definitions", params=_query_params).raise_for_status().json()
        )

    def create_build_definition(self, definition: BuildDefinitionCreate) -> BuildDefinition:
        """Create a new build definition in a project."""
        logger.debug(
            'Creating build definition "%s/%s": yaml_path="%s", yaml_repo_id="%s"',
            definition.path,
            definition.name,
            definition.process.yaml_filename,
            definition.repository.id,
        )

        logger.debug("Post request body: %s", definition.model_dump(mode="json"))
        response = self._client.post("/definitions", json=definition.model_dump(mode="json")).raise_for_status()
        serialized_response = BuildDefinition.model_validate(response.json())

        logger.debug("Build definition resource created: %s", serialized_response)
        return serialized_response

    def delete_build_definition(self, id: int) -> int:
        """Delete a build definition resource."""
        response = self._client.delete(f"/definitions/{id}").raise_for_status()
        return response.status_code



"""Module for Git Pull Request clients API."""

import logging

from ...base.clients import Client
from .models import PullRequestThread, PullRequestThreadCollection, PullRequestThreadCreate

# Get a logger for this module
logger = logging.getLogger(__name__)


class PullRequestClient(Client):
    """Represent a client to Pull Request API in Azure DevOps."""

    def find_existing_thread(self, pr_id: int, plan: str) -> PullRequestThread | None:
        """Return the existing thread ID and associated comment IDs for a project."""
        response = self._client.get(f"/{pr_id}/threads").raise_for_status()
        threads = PullRequestThreadCollection.model_validate(response.json())

        existing_thread = None
        for thread in threads:
            if not thread.deleted:
                logger.debug("Found thread: %s", thread)
                if thread.properties:
                    for property_key, property in thread.properties.items():
                        if "tfci.plan" in property_key:
                            if plan == property.value:
                                existing_thread = thread
                                logger.info("There is an existing thread for plan %s (id=%s)", plan, existing_thread)
                                logger.debug("Existing thread found: %s", existing_thread)
                                break
                    if existing_thread is not None:
                        break

        return existing_thread

    def delete_thread_comments(self, pr_id: int, thread: PullRequestThread):
        """Delete all comments for a given thread on a Pull Request."""
        logger.info("Deleting %d comments for thread with id=%s", len(thread.comments), thread.id)
        for comment in thread.comments:
            delete_request = self._client.delete(f"/{pr_id}/threads/{thread.id}/comments/{comment.id}")
            delete_request.raise_for_status()

    def new_thread(self, pr_id: int, thread: PullRequestThreadCreate):
        """Post a new thread on a pull request."""
        logger.info("Posting a thread to: %s%d", self.base_url, pr_id)
        new_thread_req = self._client.post(f"/{pr_id}/threads", json=thread.model_dump(exclude_none=True))
        new_thread_req.raise_for_status()



"""Module for Core clients API."""

from ..base.clients import Client
from .models import Project


class ProjectClient(Client):
    """Represent a client to Project API in Azure DevOps."""

    def get(self, name: str) -> Project:
        """Get a single project resource."""
        return Project.model_validate(self._client.get(name).raise_for_status().json())



"""Module for Service Hooks' Subscriptions client API."""

import logging

from ...base.clients import Client
from .models import HookSubscription, HookSubscriptionCollection, HookSubscriptionCreate

# Get a logger for this module
logger = logging.getLogger(__name__)


class HookSubscriptionClient(Client):
    """Represent a client to Git repository API in Azure DevOps."""

    def get(self, id: str) -> HookSubscription:
        """Fetch a single Service Hook's subscription."""
        logger.info("Fetching details for hook subscription with id '%s'...", id)
        resource = HookSubscription.model_validate(self._client.get(id).raise_for_status().json())
        logger.debug(resource)
        return resource

    def list(self) -> HookSubscriptionCollection:
        """List all Service Hook's available in organization."""
        return HookSubscriptionCollection.model_validate(self._client.get("").raise_for_status().json())

    def delete(self, id: str) -> None:
        """Fetch a single Service Hook's subscription."""
        logger.info("Deleting service hook subscription with id=%s...", id)
        self._client.delete(id).raise_for_status()

    def create(self, definition: HookSubscriptionCreate) -> HookSubscription:
        """Create a new Service Hook's subscription."""
        resource = HookSubscriptionCreate.model_validate(definition)
        logger.debug("Create new hook subscription: %s", definition)

        post_request = self._client.post("", json=resource.model_dump(exclude_none=True))
        post_request.raise_for_status()
        new_resource = post_request.json()
        logger.debug("Create new hook subscription status code: %s", post_request.status_code)
        logger.debug("Create new hook subscription response: %s", post_request.content)

        return HookSubscription.model_validate(new_resource)



"""Models for the Git Azure DevOps API endpoint."""

from pathlib import PurePosixPath
from typing import Annotated, Literal, Self

from pydantic import BeforeValidator, Field, HttpUrl

from ..base import validators
from ..base.models import BaseCollection, JSONModel
from ..core.models import Project
from .enums import GitObjectType, RecursionType


class GitRepository(JSONModel):
    """Model for a git repository."""

    id: str
    name: str
    project: Project
    default_branch: Annotated[str, Field(alias="defaultBranch")]
    size: int
    remote_url: Annotated[HttpUrl, Field(alias="remoteUrl")]
    web_url: Annotated[HttpUrl, Field(alias="webUrl")]
    disabled: Annotated[bool, Field(alias="isDisabled")]
    maintenance: Annotated[bool, Field(alias="isInMaintenance")]


class GitRef(JSONModel):
    """Represent a reference (branch) in a git repository."""

    name: str
    object_id: str = Field(alias="objectId")
    locked: bool | None = Field(alias="isLocked", default=None)


class GitRefCollection(BaseCollection[GitRef]):
    """Represent a collection of git repository reference."""


class GitItemDescriptor(JSONModel):
    """Represent descriptors for files and folders inside a Git repository."""

    path: Annotated[PurePosixPath, BeforeValidator(validators.posix_path)]
    version: Annotated[str, BeforeValidator(validators.git_branch_name)]
    recursion_level: Annotated[
        RecursionType,
        Field(alias="recursionLevel", default=RecursionType.full),
        BeforeValidator(RecursionType.validate),
    ] = RecursionType.full
    version_options: Annotated[
        Literal["none", "previousChange", "firstParent"], Field(alias="versionOptions", default="none")
    ] = "none"
    version_type: Annotated[Literal["branch", "tag", "commit"], Field(alias="versionType", default="branch")] = "branch"


class GitItemsBatch(JSONModel):
    """Represent the body for Git items batch requests."""

    item_descriptors: Annotated[list[GitItemDescriptor], Field(alias="itemDescriptors")]
    include_content_metadata: Annotated[bool, Field(alias="includeContentMetadata", default=False)] = False
    include_links: Annotated[bool, Field(alias="includeLinks", default=False)] = False
    latest_processed_change: Annotated[bool, Field(alias="latestProcessedChange", default=False)] = False


class GitItem(JSONModel):
    """Represent a git item (files or folders)."""

    object_id: Annotated[str, Field(alias="objectId")]
    git_object_type: Annotated[GitObjectType, BeforeValidator(GitObjectType.validate), Field(alias="gitObjectType")]
    is_folder: Annotated[bool | None, Field(alias="isFolder", default=None)] = None
    is_symlink: Annotated[bool | None, Field(alias="isSymlink", default=None)] = None
    path: Annotated[PurePosixPath, BeforeValidator(validators.posix_path)]
    content: str | None = None


class GitItemCollection(BaseCollection[GitItem]):
    """Represent a collection of git item (files or folders)."""

    def match(self, pattern: str) -> Self:
        """Return a list of items which match a given pattern."""

        def _filter(i: GitItem) -> bool:
            return i.path.match(pattern)

        filtered_results = list(filter(_filter, self))
        return type(self)(count=len(filtered_results), value=filtered_results)



"""Field validators for models."""

from pathlib import PurePosixPath, PureWindowsPath


def posix_path(value: str | PurePosixPath) -> PurePosixPath:
    """Validate the provided path is POSIX compatible."""
    if isinstance(value, PurePosixPath):
        return value
    elif isinstance(value, str):
        return PurePosixPath(value)
    else:
        raise ValueError("Cannot serialize to a path object !")


def windows_path(value: str | PureWindowsPath) -> PureWindowsPath:
    """Validate the provided path is Windows compatible."""
    if isinstance(value, PureWindowsPath):
        return value
    elif isinstance(value, str):
        return PureWindowsPath(value)
    else:
        raise ValueError("Cannot serialize to a path object !")


def git_branch_name(value: str) -> str:
    """Validate the git branch name and return the short version (without /refs/heads)."""
    return value.replace("refs/heads/", "")



"""Module for Policy Configuration client API."""

import logging

from ...base.clients import Client
from .models import PolicyConfiguration, PolicyConfigurationCreate, PolicyConfigurationUpdate

# Get a logger for this module
logger = logging.getLogger(__name__)


class PolicyConfigurationClient(Client):
    """Represent a client to Policy Configuration API in Azure DevOps."""

    def delete_policy(self, id: int) -> int:
        """Delete a policy configuration by ID."""
        logger.info("Deleting policy configuration ID %s...", id)
        response = self._client.delete(str(id)).raise_for_status()
        return response.status_code

    def create_build_policy(self, definition: PolicyConfigurationCreate) -> PolicyConfiguration:
        """Create a new build policy for a branch."""
        logger.debug("Creating a new build policy: %s", definition)
        response = self._client.post("", json=definition.model_dump(mode="json")).raise_for_status()
        return PolicyConfiguration.model_validate(response.json())

    def update_build_policy(self, id: int, definition: PolicyConfigurationUpdate) -> PolicyConfiguration:
        """Update an existing build policy by ID for a branch."""
        logger.debug("Updating a new build policy with id=%d: %s", id, definition)
        response = self._client.put(f"{id}", json=definition.model_dump(mode="json")).raise_for_status()
        return PolicyConfiguration.model_validate(response.json())



"""Module for Git client API."""

import logging

from pathlib import PurePosixPath
from typing import Any

from ..base.clients import Client
from ..policy.configurations.models import PolicyConfigurationCollection
from .models import GitItem, GitItemCollection, GitItemDescriptor, GitItemsBatch, GitRefCollection, GitRepository

# Get a logger for this module
logger = logging.getLogger(__name__)


class GitRepositoryClient(Client):
    """Represent a client to Git repository API in Azure DevOps."""

    def get(self, repository: str) -> GitRepository:
        """Fetch a single Git repository."""
        logger.info("Fetching details for repository '%s'...", repository)
        return GitRepository.model_validate(self._client.get(repository).raise_for_status().json())

    def get_refs(self, repository: str, branch_startswith: str | None = None) -> GitRefCollection:
        """Fetch all refs for a single Git repository."""
        req_params = {}
        if branch_startswith:
            req_params.update({"filter": branch_startswith.replace("refs/", "")})
        return GitRefCollection.model_validate(
            self._client.get(f"/{repository}/refs", params=req_params).raise_for_status().json()
        )

    def get_items_batch(self, repository: str, item_descriptors: GitItemsBatch) -> GitItemCollection:
        """Git a batch of items (files and folders) from a repository."""
        request_body = item_descriptors.model_dump(mode="json")
        logger.debug("Items batch Body: %s", request_body)
        response = self._client.post(f"/{repository}/itemsbatch", json=request_body).raise_for_status()
        logger.debug("Items batch POST: %s", response.url)
        response_dict = response.json()
        _response = {"count": len(response_dict["value"][0]), "value": response_dict["value"][0]}
        serialized_response = GitItemCollection.model_validate(_response)
        logger.debug("Items batch response: %s", serialized_response)
        return serialized_response

    def get_item(self, repository: str, path: str | PurePosixPath, branch: str | None = None) -> GitItem:
        """Get an item (files, folders and symlinks) from a git repository."""
        if isinstance(path, str):
            _path = PurePosixPath(path)
        elif isinstance(path, PurePosixPath):
            _path = path
        else:
            raise ValueError("The path argument can be either a str or PurePosixPath instance !")

        req_params = {
            "path": _path.as_posix(),
            "download": "false",
            "includeContent": "true",
            "$format": "json",
        }

        if branch:
            req_params["versionDescriptor.version"] = branch

        response = self._client.get(f"/{repository}/items", params=req_params).raise_for_status()
        logger.debug("get_item: %s\nResponse: %s", response.url, response.json())
        return GitItem.model_validate(response.json())

    def list_files(self, repository: str, branch: str, path: PurePosixPath | str = "/") -> GitItemCollection:
        """Return files in a given path of a Git repository."""
        if not isinstance(path, PurePosixPath):
            _path = PurePosixPath(path)
        else:
            _path = path

        return self.get_items_batch(
            repository=repository,
            item_descriptors=GitItemsBatch(item_descriptors=[GitItemDescriptor(path=_path, version=branch)]),
        )

    def is_branch_locked(self, repository: str, branch: str) -> bool:
        """Check for branch lock status."""
        ref_info = self.get_refs(repository=repository, branch_startswith=branch)[0]
        logger.debug("Checking branch lock status for '%s@%s': %s", repository, branch, ref_info)
        return ref_info.locked if ref_info.locked else False

    def lock_branch_toggle(self, repository: str, branch: str, locked: bool) -> dict[str, Any]:
        """Lock / Unlock a branch on a single Git repository."""
        req_params = {}
        branch_status = "locked" if locked else "unlocked"
        if branch:
            req_params.update({"filter": branch.replace("refs/", "")})
        logger.info("Branch '%s' for '%s' repository is now %s.", branch, repository, branch_status)
        response = self._client.patch(
            f"/{repository}/refs", json={"isLocked": locked}, params=req_params
        ).raise_for_status()
        return response.json()


class GitPolicyConfigurationClient(Client):
    """Represent a client to Git Repository Policy Configuration API in Azure DevOps."""

    def get_policies_for_ref(
        self, repository_id: str, ref_name: str, policy_type: str
    ) -> PolicyConfigurationCollection:
        """Get all policies for a branch of a repository."""
        response = self._client.get(
            "", params={"repositoryId": repository_id, "refName": ref_name, "policyType": policy_type}
        ).raise_for_status()
        logger.debug("Request URL: %s\nResponse: %s", response.url, response.content)
        return PolicyConfigurationCollection.model_validate(response.json())

    def get_build_policies_for_ref(self, repository_id: str, ref_name: str) -> PolicyConfigurationCollection:
        """Get all build policies for a branch of a repository."""
        policies = self.get_policies_for_ref(
            repository_id=repository_id,
            ref_name=ref_name,
            policy_type="0609b952-1397-4640-95ec-e00a01b2c241",
        )
        return policies



"""Enumerations for Service Hooks' Subscriptions types."""

from ...base.enums import StrIntEnum


class SubscriptionStatus(StrIntEnum):
    """
    Enum for available hook subscriptions states types.

    Reference: https://learn.microsoft.com/en-us/javascript/api/azure-devops-extension-api/subscriptionstatus
    """

    enabled = 0
    onProbation = 10
    disabledByUser = 20
    disabledBySystem = 30
    disabledByInactiveIdentity = 40



"""Models for the Builds Azure DevOps API endpoint."""

from pathlib import PurePosixPath, PureWindowsPath
from typing import Annotated, Literal, Self

from pydantic import BeforeValidator, Field, HttpUrl, computed_field

from ..base import validators
from ..base.models import BaseCollection, JSONModel
from ..core.models import Properties


class BuildProcess(JSONModel):
    """Represents a build process resource."""

    type: int = 2
    yaml_filename: Annotated[PurePosixPath, BeforeValidator(validators.posix_path), Field(alias="yamlFilename")]


class BuildRepository(JSONModel):
    """Represents a repository used to create a build definition."""

    id: str
    type: Literal["TfsGit"] = "TfsGit"
    default_branch: Annotated[str | None, Field(alias="defaultBranch", default=None)] = None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def default_branch_name(self) -> str | None:
        """Return the short name for the default branch (without refs/heads/)."""
        if self.default_branch:
            return validators.git_branch_name(self.default_branch)
        else:
            return None


class BuildDefinitionBase(JSONModel):
    """The base class of a build definition."""

    id: int
    revision: int
    name: str
    path: Annotated[PureWindowsPath, BeforeValidator(validators.windows_path)]
    url: HttpUrl
    properties: dict[str, Properties] | None = None


class BuildDefinitionCreate(JSONModel):
    """Represents a request to create a new build definition."""

    name: str
    path: Annotated[PureWindowsPath, BeforeValidator(validators.windows_path)]
    type: Literal["xaml", "build"] = "build"
    process: BuildProcess
    repository: BuildRepository
    properties: dict[str, Properties] | None = None


class BuildDefinition(BuildDefinitionBase):
    """Represents a build definition resource."""

    process: BuildProcess
    repository: BuildRepository

    @computed_field  # type: ignore[prop-decorator]
    @property
    def fullname(self) -> PurePosixPath:
        """Return the fullname including path."""
        return PurePosixPath(self.path.joinpath(self.name))


class BuildDefinitionCollection(BaseCollection[BuildDefinitionBase]):
    """Represents a collection of build definition resources."""

    def startswith(self, pattern: str) -> Self:
        """Filter build definitions that have a name starting with `pattern`."""

        def _filter(p: BuildDefinitionBase) -> bool:
            return p.name.startswith(pattern)

        filtered_results = list(filter(_filter, self))
        return type(self)(count=len(filtered_results), value=filtered_results)

    def from_folder(self, pattern: str) -> Self:
        """Filter build definitions that are contained in a folder."""

        def _filter(p: BuildDefinitionBase) -> bool:
            return p.path.match(pattern)

        filtered_results = list(filter(_filter, self))
        return type(self)(count=len(filtered_results), value=filtered_results)

    def get(self, name: str) -> BuildDefinitionBase | None:
        """Get a build definition by `name` or None if not found."""
        for item in self:
            if item.name == name:
                return item
        return None
